#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>
/* LOCAL INCLUDES */
#include	<malloclib.h>
#include	<defs.h>
#include	<types.h>
#include	<lum.h>
#include	<errexit.h>
#include	<ppm.h>

void P_ppm_make_header(PPMheader *hp)
{
  if(((*hp) = (PPMheader)malloc(sizeof(PPMheader_struct)))==NULL)
    errexit("P_ppm_make_header: malloc: %s\n",strerror(errno));
}


void P_ppm_free_header(PPMheader *hp)
{
  free((*hp));
}


void P_ppm_copy_header(PPMheader ihp, PPMheader ohp)
{
  ohp->rows = ihp->rows;
  ohp->cols = ihp->cols;
  ohp->format = ihp->format;
  ohp->maxval = ihp->maxval;
}


void P_ppm_read_header(FILE *fp, PPMheader hp)
{
	int	ich1, ich2;

  if((ich1 = getc(fp))==EOF)
    errexit("P_ppm_read_header: EOF / read error reading magic number\n");
  if((ich2 = getc(fp))==EOF)
    errexit("P_ppm_read_header: EOF / read error reading magic number\n");
  hp->format = ich1 * 256 + ich2;

  switch(PPM_FORMAT_TYPE(hp->format)) {
    case PPM_TYPE:
      hp->cols = P_pnm_getint(fp);
      hp->rows = P_pnm_getint(fp);
      hp->maxval = P_pnm_getint(fp);
      if(hp->maxval > PPM_MAXMAXVAL)
        errexit("P_ppm_read_header: maxval is too large - try reconfiguring with PGM_BIGGRAYS\n");
      break;
    case PGM_TYPE:
      hp->cols = P_pnm_getint(fp);
      hp->rows = P_pnm_getint(fp);
      hp->maxval = P_pnm_getint(fp);
      if(hp->maxval > PGM_MAXMAXVAL)
        errexit("P_ppm_read_header: maxval is too large - try reconfiguring with PGM_BIGGRAYS\n" );
      break;
    case PBM_TYPE:
      hp->cols = P_pnm_getint(fp);
      hp->rows = P_pnm_getint(fp);
      hp->maxval = 1;
      break;
    default:
      errexit("P_ppm_read_header: bad magic number - not a pgm or pbm file\n");
  }
}


void P_ppm_get_rows(PPMheader hp, int *rows)
{
  *rows = hp->rows;
}


void P_ppm_get_cols(PPMheader hp, int *cols)
{
  *cols = hp->cols;
}


void P_ppm_get_format(PPMheader hp, int *format)
{
  *format = hp->format;
}


void P_ppm_get_maxval(PPMheader hp, gray *maxval)
{
  *maxval = hp->maxval;
}


void P_ppm_set_rows(PPMheader hp, int rows)
{
  hp->rows = rows;
}


void P_ppm_set_cols(PPMheader hp, int cols)
{
  hp->cols = cols;
}


void P_ppm_set_format(PPMheader hp, int format)
{
  hp->format = format;
}


void P_ppm_set_maxval(PPMheader hp, gray maxval)
{
  hp->maxval = maxval;
}


void P_ppm_write_header(FILE *fp, PPMheader hp)
{
  fprintf(fp,"%c%c\n%d %d\n%d\n",PPM_MAGIC1,RPPM_MAGIC2,
	hp->cols,hp->rows,hp->maxval);
}


void P_ppm_read_larr(FILE *fp,ulong *larr,int r,int c,int format,gray maxv)
{
	int		i,j;
	register int	bitshift;
	register gray	g, item;
	register bit	b;
	register uchar	blue,green,red,alpha=0xff;

  for(i=0;i<r;++i) {
    bitshift = -1;
    for(j=0;j<c;j++) {
      switch(format) {
        case PPM_FORMAT:
          red = P_pnm_getint(fp);
          green = P_pnm_getint(fp);
          blue = P_pnm_getint(fp);
          break;
        case RPPM_FORMAT:
          red = P_pnm_getrawbyte(fp);
          green = P_pnm_getrawbyte(fp);
          blue = P_pnm_getrawbyte(fp);
          break;
        case PGM_FORMAT:
          g = P_pnm_getint(fp);
          blue = green = red = (uchar)g;
          break;
        case RPGM_FORMAT:
          g = P_pnm_getrawbyte(fp);
          blue = green = red = (uchar)g;
          break;
        case PBM_FORMAT:
        case RPBM_FORMAT:
          switch(format) {
            case PBM_FORMAT:
              b = P_pnm_getbit(fp);
              break;
            case RPBM_FORMAT:
              if(bitshift == -1) {
                item = P_pnm_getrawbyte(fp);
                bitshift = 7;
              }
              b = (item >> bitshift) & 1;
              --bitshift;
              break;
            default:
              errexit("P_ppm_read_larr: unknown PBM format\n");
          }
          g = (b==PBM_WHITE) ? maxv : 0;
          blue = green = red = (uchar)g;
          break;
        default:
          errexit("P_ppm_read_larr: unknown PGM format\n");
      }
      /* store in ABGR format (obsolete)
      larr[M2A(i,j,r,c)] = ((alpha<<24) | (blue<<16) | (green<<8) | red);
      */
      /* store in RGBA format */
      larr[M2A(i,j,r,c)] = ((red<<24) | (green<<16) | (blue<<8) | alpha);
    }
  }
}


void P_ppm_write_larr(FILE *fp,ulong *larr,int r,int c,int format,gray maxv)
{
	int		i,j,charcount;
	register gray	g;
	register uchar	blue,green,red;

#ifdef PBMPLUS_RAWBITS
  for(i=0;i<r;i++) {
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R(larr[M2A(i,j,r,c)]);
      green	= (uchar)LONG_G(larr[M2A(i,j,r,c)]);
      blue	= (uchar)LONG_B(larr[M2A(i,j,r,c)]);
      putc(red,fp);
      putc(green,fp);
      putc(blue,fp);
    }
  }
#else
  for(i=0;i<r;i++) {
    charcount = 0;
    for(j=0;j<c;j++) {
      red	= (uchar)LONG_R(larr[M2A(i,j,r,c)]);
      green	= (uchar)LONG_G(larr[M2A(i,j,r,c)]);
      blue	= (uchar)LONG_B(larr[M2A(i,j,r,c)]);
      P_pnm_putus((ushort)red,fp);
      P_pnm_putus((ushort)green,fp);
      P_pnm_putus((ushort)blue,fp);
      if(charcount >= 65) {
        putc('\n',fp);
        charcount = 0;
      }
      else if(charcount > 0) {
        putc(' ',fp);
        putc(' ',fp);
        charcount += 2;
      }
      P_pnm_putus((ushort)red,fp);
      putc(' ',fp);
      P_pnm_putus((ushort)green,fp);
      putc(' ',fp);
      P_pnm_putus((ushort)blue,fp);
      putc(' ',fp);
      charcount += 11;
    }
    if(charcount > 0)
      putc('\n',fp);
  }
#endif
}
