/* 
 * Copyright (C) 1998 Janne Lf <jlof@mail.student.oulu.fi>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <math.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtkgl/gtkglarea.h>
#include "types.h"

typedef unsigned int	uint;
typedef unsigned long	ulong;
typedef unsigned short	ushort;

#include "ppm.h"

char help_text[] = "Usage: viewppm [OPTION]... [FILE]...\n"
                   "View PPM images.\n"
                   "\n"
                   "Options:\n"
                   "  --help                            display help\n"
                   "\n"
                   "In the program:\n"
                   "  Mouse button 1 + drag             spin (virtual trackball)\n"
                   "  Mouse button 2 + drag             zoom\n"
                   "  Mouse button 3                    popup menu\n"
                   "\n";

typedef struct {
	PPMheader	ihp;
	ulong		*larr;
	int		do_init;
} PPMImg_struct, *PPMImg;

gint show_ppmimage(char *ppmimage_name);
void select_ppmimage();
void ppmimg_free(PPMImg *ppmimg);

void initgl(void)
{
  /* specify clear color (black) */
  glClearColor(0.0,0.0,0.0,0.0);

  /* turn on depth testing (z-buffering) */
  glEnable(GL_DEPTH_TEST);
  glClearDepth(1.0);

  /* turn on blending (for transparency) */
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
}

gint glarea_expose(GtkWidget *widget, GdkEventExpose *event)
{
  GtkGLArea *glarea = GTK_GL_AREA(widget);

  PPMImg ppmimg = (PPMImg)gtk_object_get_data(GTK_OBJECT(widget),"ppmimg");

  if(!ppmimg->ihp || !ppmimg->larr) {
    return FALSE;
  }

  /* draw only last expose */
  if (event->count > 0) {
    return TRUE;
  }

  /* OpenGL calls can be done only if begingl returns true */
  if (gtk_gl_area_begingl(glarea)) {
    /* basic initialization */
    if (ppmimg->do_init == TRUE) {
      initgl();
      ppmimg->do_init = FALSE;
    }

    /* view */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0,ppmimg->ihp->cols,0.0,ppmimg->ihp->rows);

#ifndef MESA
glDrawBuffer(GL_FRONT_AND_BACK);
#endif /* !MESA */
    glClearColor(0.0,0.0,0.0,0.0);
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* draw object */
    glRasterPos2i(0,ppmimg->ihp->rows);
    glPixelZoom(1.0,-1.0);
    glDrawPixels(ppmimg->ihp->cols,ppmimg->ihp->rows,GL_RGBA,GL_UNSIGNED_BYTE,ppmimg->larr);
  
    /* opengl rendering done for now */
    gtk_gl_area_endgl(glarea);
  }
  /* swap backbuffer to front */
  gtk_gl_area_swapbuffers(glarea);

  return TRUE;
}

gint glarea_configure(GtkWidget *widget, GdkEventConfigure *event)
{
  /* OpenGL calls can be done only if begingl returns true */
  if (gtk_gl_area_begingl(GTK_GL_AREA(widget))) {
    glViewport(0,0, widget->allocation.width, widget->allocation.height);

    /* end opengl calls by calling endgl */
    gtk_gl_area_endgl(GTK_GL_AREA(widget));
  }
  return TRUE;
}

gint glarea_destroy(GtkWidget *widget)
{
  PPMImg ppmimg = (PPMImg)gtk_object_get_data(GTK_OBJECT(widget),"ppmimg");

  /* delete image */
  ppmimg_free(&ppmimg);

  return TRUE;
}

gint glarea_button_press(GtkWidget *widget, GdkEventButton *event)
{
  PPMImg ppmimg = (PPMImg)gtk_object_get_data(GTK_OBJECT(widget),"ppmimg");

  if (event->button == 1) {
    /* beginning of drag, reset mouse position */
/*
    info->beginx = event->x;
    info->beginy = event->y;
*/
    return TRUE;
  }
  return FALSE;
}

gint popup_menu_handler(GtkWidget *widget, GdkEventButton *event)
{
  if (event->button == 3) {
    gtk_menu_popup(GTK_MENU(widget), NULL, NULL, NULL, NULL,
		   event->button, event->time);
    return TRUE;
  }
  return FALSE;
}
void popup_menu_detacher(GtkWidget *attach_widget,GtkMenu *menu)
{
}
void create_popup_menu(GtkWidget *widget)
{
  GtkWidget *menu          = gtk_menu_new();
  GtkWidget *open_item     = gtk_menu_item_new_with_label("Open");
  GtkWidget *quit_item     = gtk_menu_item_new_with_label("Quit");
  GtkWidget *quit_all_item = gtk_menu_item_new_with_label("Quit All");

  gtk_menu_append(GTK_MENU(menu), open_item);
  gtk_menu_append(GTK_MENU(menu), quit_item);
  gtk_menu_append(GTK_MENU(menu), quit_all_item);

  gtk_widget_show(open_item);
  gtk_widget_show(quit_item);
  gtk_widget_show(quit_all_item);

  gtk_menu_attach_to_widget(GTK_MENU(menu),GTK_WIDGET(widget),popup_menu_detacher);
  gtk_signal_connect_object(GTK_OBJECT(widget), "destroy",
			    GTK_SIGNAL_FUNC(gtk_menu_detach), GTK_OBJECT(menu));

 gtk_signal_connect_object(GTK_OBJECT(widget), "button_press_event",
			    GTK_SIGNAL_FUNC(popup_menu_handler), GTK_OBJECT(menu));


  gtk_signal_connect(GTK_OBJECT(open_item), "activate",
		     GTK_SIGNAL_FUNC(select_ppmimage), NULL);

  gtk_signal_connect_object(GTK_OBJECT(quit_item), "activate",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(widget));

  gtk_signal_connect(GTK_OBJECT(quit_all_item), "activate",
		     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
}



gint window_count = 0; /* number of windows on screen */
gint window_destroy(GtkWidget *widget)
{
  /* if this was last window quit */
  if (--window_count == 0)
    gtk_main_quit();
  return TRUE;
}



gint show_ppmimage(char *ppmimage_name)
{
  GtkWidget	*window, *frame, *glarea;
  PPMImg	ppmimg;
  PPMheader	ihp;
  ulong		*larr=NULL;

  FILE		*fp;

  /* read ppm image */
  if(!(fp = fopen(ppmimage_name,"r"))) {
    g_print("can't open file %s\n", ppmimage_name);
    return FALSE;
  }
  P_ppm_make_header(&ihp);
  P_ppm_read_header(fp,ihp);
  if((larr = (ulong *)malloc(ihp->rows*ihp->cols*sizeof(ulong)))==NULL) {
    g_print("not enough memory\n");
    return FALSE;
  }
  P_ppm_read_larr(fp,larr,ihp->rows,ihp->cols,ihp->format,ihp->maxval);

  ppmimg = (PPMImg)malloc(sizeof(PPMImg_struct));
  ppmimg->do_init = FALSE;
  ppmimg->ihp = ihp;
  ppmimg->larr = larr;

  /* create aspect frame */
  frame = gtk_aspect_frame_new(NULL, 0.5, 0.5, ihp->cols/ihp->rows, FALSE);

  /* create new OpenGL widget with GTK's window mode; analogous to
     glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  mode(GL_DOUBLE | GL_RGBA | GL_DEPTH);
  */

  glarea = gtk_gl_area_new_vargs(NULL, /* no sharing */
				 GDK_GL_DOUBLEBUFFER,
				 GDK_GL_RGBA,
				 GDK_GL_DEPTH_SIZE,1,
				 GDK_GL_NONE);  /* last argument must be GDK_GL_NONE */
  if (glarea == NULL) {
    ppmimg_free(&ppmimg);
    g_print("Can't create GtkGLArea widget\n");
    return FALSE;
  }
  /* set up events and signals for OpenGL widget */
  gtk_widget_set_events(glarea,
			GDK_EXPOSURE_MASK|
                        GDK_BUTTON_PRESS_MASK|
                        GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect (GTK_OBJECT(glarea), "expose_event",
		      GTK_SIGNAL_FUNC(glarea_expose), NULL);
  gtk_signal_connect (GTK_OBJECT(glarea), "button_press_event",
                      GTK_SIGNAL_FUNC(glarea_button_press), NULL);
  gtk_signal_connect (GTK_OBJECT(glarea), "configure_event",
		      GTK_SIGNAL_FUNC(glarea_configure), NULL);
  gtk_signal_connect (GTK_OBJECT(glarea), "destroy",
		      GTK_SIGNAL_FUNC (glarea_destroy), NULL);

  gtk_widget_set_usize(glarea, ihp->cols, ihp->rows); /* minimum size */

  gtk_object_set_data(GTK_OBJECT(glarea), "ppmimg", ppmimg);

  /* create new top level window */
  window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), ppmimage_name);
  gtk_container_border_width(GTK_CONTAINER(window), 10);
  create_popup_menu(window); /* add popup menu to window */
  gtk_signal_connect (GTK_OBJECT(window), "destroy",
		      GTK_SIGNAL_FUNC(window_destroy), NULL);
  window_count++;

  /* destroy this window when exiting from gtk_main() */
  gtk_quit_add_destroy(1, GTK_OBJECT(window));


  /* put glarea into window and show it all */
  gtk_container_add(GTK_CONTAINER(window), frame);
  gtk_container_add(GTK_CONTAINER(frame),glarea);
  gtk_widget_show(glarea);
  gtk_widget_show(frame);
  gtk_widget_show(window);

  return TRUE;
}


void ppmimg_free(PPMImg *ppmimg)
{
  free((*ppmimg)->ihp);
  free((*ppmimg)->larr);
  free((*ppmimg));
  (*ppmimg) = NULL;
}


gint filew_ok(GtkWidget *widget, GtkWidget *filew)
{
  if (show_ppmimage(gtk_file_selection_get_filename(GTK_FILE_SELECTION(filew))) == TRUE)
    gtk_widget_destroy(filew);
  return TRUE;
}

void select_ppmimage()
{
  GtkWidget *filew = gtk_file_selection_new("Select PPM image");

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION (filew)->ok_button), "clicked",
		     GTK_SIGNAL_FUNC(filew_ok), filew);

  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filew)->cancel_button), "clicked",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filew));

  gtk_signal_connect (GTK_OBJECT(filew), "destroy",
		      GTK_SIGNAL_FUNC(window_destroy), NULL);
  window_count++;

  gtk_widget_show(filew);
}





int main (int argc, char **argv)
{
	int	num_events;

  /* initialize gtk */
  gtk_init( &argc, &argv );

  /* Check if OpenGL (GLX extension) is supported. */
  if (gdk_gl_query() == FALSE) {
    g_print("OpenGL not supported\n");
    return 0;
  }

  /* help? */
  if (argc >= 2 && strcmp(argv[1],"--help")==0) {
    g_print("%s", help_text);
    return 0;
  }

  if (argc == 1) {
    /* no filenames, show filerequester */
    select_ppmimage();
  } else {
    /* show requested files */
    int i;
    for (i=1; i<argc; i++)
      show_ppmimage(argv[i]);
  }

/* old, "loss of control" flow-of-control hand-off to gtk
  if (window_count > 0)
    gtk_main();
*/
  while(1 && window_count) {
    if((num_events = gtk_events_pending()))
      gtk_main_iteration();    
    fprintf(stderr,"in loop\n");
  }

  return 0;
}
